/**************************************************************************
Copyright (C) 2002-2004, Marco Barisione <marco@barisione.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
**************************************************************************/

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <stdlib.h>
#include "version.h"
#include "resource.h"

HINSTANCE hInst;

#define BUFF_SIZE 4096


void text_out(HWND hwnd, TCHAR * txt){
    SetDlgItemText(hwnd, IDC_OUTPUT, txt);
}


void text_out_file(HWND hwnd, char * filename, TCHAR * alt_text){
    FILE * file;
    size_t len;
    TCHAR * txt;
    if (file = fopen(filename, "rb")){
        fseek(file, 0, SEEK_END);
        len = ftell(file);
        txt = malloc(sizeof(TCHAR) * (len + 1));
        fseek(file, 0, SEEK_SET);
        fread(txt, len, 1, file);
        txt[len] = '\0';
        text_out(hwnd, txt);
        free(txt);
        fclose(file);
    }
    else
        text_out(hwnd, alt_text);
}


TCHAR * convert_eols(TCHAR * unix_text){
    TCHAR * ret;
    TCHAR * tp = unix_text;
    size_t len = 0;
    while (*tp != '\0'){
        ++len;
        if (*tp == '\n')
            ++len;
        ++tp;
    }
    ret = malloc(sizeof(TCHAR) * (len + 1));
    ret[len] = '\0';
    tp = ret;
    while (*unix_text != '\0'){
        if (*unix_text == '\n'){
            *tp = '\r';
            ++tp;
        }
        *tp = *unix_text;
        ++tp;
        ++unix_text;
    }
    return ret;
}


void about(HWND hwnd){
    text_out(hwnd,
             _T("WinWhois (ver. ") _T(VERSION) _T(")\r\n")
             _T("Copyright  2002-2006 Marco Barisione ")
             _T("<marco@barisione.org>\r\n")
             _T("http://www.barisione.org/apps.html#winwhois\r\n\r\n")
             _T("whois (ver. ") _T(W_VERSION) _T(")\r\n")
             _T("Copyright  1999-2003 Marco d'Itri ")
             _T("<md@linux.it>\r\n")
             _T("http://www.linux.it/~md/software/"));
}


BOOL call_whois(HWND hwnd, TCHAR * txt){
    int buff_size = 0;
    int free_buff = 0;
    int buff_index = 0;
    TCHAR * buff = NULL;
    TCHAR * new_buff;
    DWORD bytes_read;
    TCHAR * cmd;
    SECURITY_ATTRIBUTES sec_attr;
    BOOL ret;
    HANDLE child_stdout_read, child_stdout_write, child_stdout_read_dup,
           prev_stdout, prev_stderr;
    PROCESS_INFORMATION proc_info;
    STARTUPINFO startup_info;
    
    sec_attr.nLength = sizeof(SECURITY_ATTRIBUTES);
    sec_attr.bInheritHandle = TRUE;
    sec_attr.lpSecurityDescriptor = NULL;

    prev_stdout = GetStdHandle(STD_OUTPUT_HANDLE);
    prev_stderr = GetStdHandle(STD_ERROR_HANDLE);

    if (!CreatePipe(&child_stdout_read, &child_stdout_write, &sec_attr, 0))
        return FALSE;

    if (!SetStdHandle(STD_OUTPUT_HANDLE, child_stdout_write) ||
        !SetStdHandle(STD_ERROR_HANDLE, child_stdout_write))
        return FALSE;

    ret = DuplicateHandle(GetCurrentProcess(), child_stdout_read,
                          GetCurrentProcess(), &child_stdout_read_dup,
                          0, FALSE, DUPLICATE_SAME_ACCESS);
    if(!ret)
        return FALSE;
    CloseHandle(child_stdout_read);

    ZeroMemory(&startup_info, sizeof(STARTUPINFO));
    startup_info.cb = sizeof(STARTUPINFO);
    startup_info.hStdInput = NULL;
    startup_info.hStdOutput = child_stdout_write;
    startup_info.hStdError = child_stdout_write;
    startup_info.dwFlags = STARTF_USESTDHANDLES;

    cmd = malloc(sizeof(TCHAR) * (12 /* ".\\whois.exe " */ + lstrlen(txt)
                                  + 1 /* '\0' */));
    wsprintf(cmd, _T("%s %s"), _T(".\\whois.exe"), txt);
    if (!CreateProcess(NULL, cmd, NULL, NULL, TRUE, DETACHED_PROCESS,
        NULL, NULL, &startup_info, &proc_info))
        return FALSE;
    free(cmd);

    SetStdHandle(STD_OUTPUT_HANDLE, prev_stdout);
    SetStdHandle(STD_ERROR_HANDLE, prev_stderr);

    CloseHandle(child_stdout_write);

    for (;;){
        if (!free_buff){
            buff_size += BUFF_SIZE;
            buff = realloc(buff, buff_size + 1);
            free_buff = BUFF_SIZE;
        }
        ret = ReadFile(child_stdout_read_dup, &buff[buff_index],
                       free_buff, &bytes_read, NULL);
        if (bytes_read == 0 || !ret){
            buff[buff_index + bytes_read] = '\0';
            break;
        }
        free_buff -= bytes_read;
        buff_index += bytes_read;
    }

    CloseHandle(child_stdout_read_dup);
    new_buff = convert_eols(buff);
    text_out(hwnd, new_buff);
    free(buff);
    free(new_buff);
    return TRUE;
}


void on_size(HWND hwnd){
    const int space = 10;
    RECT win_rect, btn_rect, label_rect;
    HWND label = GetDlgItem(hwnd, IDC_LABEL);
    HWND text = GetDlgItem(hwnd, IDC_CMD);
    HWND button = GetDlgItem(hwnd, IDC_OK);
    HWND output = GetDlgItem(hwnd, IDC_OUTPUT);
    INT btn_left;
    INT btn_height;
    GetClientRect(hwnd, &win_rect);
    /* label */
    SetWindowPos(label, 0, space, space, 0, 0,
                 SWP_NOSIZE | SWP_NOZORDER);
    /* button */
    GetWindowRect(button, &btn_rect);
    btn_left = win_rect.right - (btn_rect.right - btn_rect.left) - space;
    SetWindowPos(button, 0, btn_left, space, 0, 0,
                 SWP_NOSIZE | SWP_NOZORDER);
    /* text */
    btn_height = btn_rect.bottom - btn_rect.top;
    GetWindowRect(label, &label_rect);
    SetWindowPos(text, 0, label_rect.right + space, space,
                 btn_left - label_rect.right - space * 2,
                 btn_height, SWP_NOZORDER);
    /* output */
    GetWindowRect(button, &btn_rect);
    SetWindowPos(output, 0, space, space * 2 + btn_height,
                 win_rect.right - space * 2,
                 win_rect.bottom - btn_height - space * 3,
                 SWP_NOZORDER);
}


void set_title(HWND hwnd, BOOL wait){
    TCHAR * txt;
    if (wait)
        txt = _T("Please wait...");
    else
        txt = _T("WinWhois - version ") _T(VERSION);
    SetWindowText(hwnd, txt);
}


void init_dialog(HWND hDlg){
    HICON icon;
    HFONT fixed_font;

    icon= LoadIcon(hInst, MAKEINTRESOURCE(IDI_APP_ICO));
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_BIG, (LPARAM)icon);
    SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_SMALL, (LPARAM)icon);

    fixed_font = (HFONT)GetStockObject(ANSI_FIXED_FONT);
    SendDlgItemMessage(hDlg, IDC_OUTPUT, WM_SETFONT,
                       (WPARAM)fixed_font, 0);

    on_size(hDlg);
    about(hDlg);
    set_title(hDlg, FALSE);
}


BOOL CALLBACK DlgProc(HWND hDlg,
                      UINT uMsg,
                      WPARAM wParam,
                      LPARAM lParam){
    TCHAR * txt;
    HWND txt_ctrl;
    int len;
    switch (uMsg){
        case WM_INITDIALOG:
            init_dialog(hDlg);
            break;
        case WM_SIZE:
            on_size(hDlg);
            break;
        case WM_COMMAND:
            switch (LOWORD(wParam)){
                case IDCANCEL:
                    EndDialog(hDlg, 0);
                    break;
                case IDC_OK:
                    set_title(hDlg, TRUE);
                    txt_ctrl = GetDlgItem(hDlg, IDC_CMD);
                    len = GetWindowTextLength(txt_ctrl);
                    if (len){
                        txt = malloc(sizeof(TCHAR) * (len + 1));
                        GetWindowText(txt_ctrl, txt, len + 1);
                        if (!call_whois(hDlg, txt))
                            MessageBox(hDlg, _T("Cannot launch whois."),
                                       _T("Error"), MB_ICONERROR);
                        free(txt);
                    }
                    set_title(hDlg, FALSE);
                    break;
                case IDM_ABOUT:
                    about(hDlg);
                    break;
                case IDM_HELP_WHOIS:
                    if (!call_whois(hDlg, _T("")))
                        MessageBox(hDlg, _T("Cannot dispaly help."),
                                   _T("Error"), MB_ICONERROR); 
                    break;
                case IDM_LICENSE:
                    text_out_file(hDlg, "license.txt",
                                  _T("\"license.txt\" not found.\r\n")
                                  _T("You can find the license at ")
                                  _T("http://www.gnu.org/copyleft/gpl.html"));
                    break;
            }
            break;
    }
    return FALSE;
}


INT APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     INT       nCmdShow){
    TCHAR dir[MAX_PATH];
    int i;
    hInst = hInstance;
    GetModuleFileName(NULL, dir, MAX_PATH);
    for (i = lstrlen(dir); i >= 0; --i){
        if (dir[i] == '\\' || dir[i] == ':'){
            dir[i] = '\0';
            break;
        }
    }
    SetCurrentDirectory(dir);
    return DialogBox(hInst, MAKEINTRESOURCE(IDD_DLG), NULL, DlgProc) == -1;
}
